import importlib
import inspect

from typing import Callable


def get_class(module_name: str, module_path: str) -> tuple[str, Callable]:
    """Retrieves the correct class name of a module"""

    # Import the module
    module = importlib.import_module(module_name, module_path)

    for class_name, class_def in inspect.getmembers(module, inspect.isclass):
        if class_name.lower() == module_name.lower().split(".")[-1]:
            return class_name, class_def

    raise ModuleNotFoundError(
        f"Couldnt find corresponding class {module_name} in {module_path}."
    )



import os

from typing import Callable

from pyDataverse.api import NativeApi
from easyDataverse.tools.utils import get_class
from easyDataverse.core.exceptions import MissingCredentialsException, MissingURLException


def download_from_dataverse(cls, doi: str, lib_name: str):
    """Downloads a dataset from a Dataverse instance and initializes a Dataset object.

    Args:
        cls (Dataset): Empty dataset to which will be written.
        lib_name (str): INTERNAL. Used to derive the underlying generated API.
    """

    # Get environment variables
    try:
        DATAVERSE_URL = os.environ["DATAVERSE_URL"]

    except KeyError:
        raise MissingURLException

    try:
        API_TOKEN = os.environ["DATAVERSE_API_TOKEN"]
    except KeyError:
        raise MissingCredentialsException

    # Intialize the pyDataverse instance to fetch the dataset
    api = NativeApi(DATAVERSE_URL, API_TOKEN)

    # Download the dataset and retrieve the field data
    dataset = api.get_dataset(doi)
    setattr(cls, "p_id", doi)
    json_data = dataset.json()["data"]["latestVersion"]["metadataBlocks"]

    # Build new mapping
    data = {}

    for block_name, block in json_data.items():

        module_path = f".metadatablocks.{block_name}"

        _, module = get_class(module_path, lib_name)

        fields = block["fields"]
        data = {}

        for field in fields:
            parser_fun = field_parser_factory(field["typeClass"])
            data.update(parser_fun(field, module))

        # Initialize the module
        cls.add_metadatablock(module.parse_obj(data))

    return cls


def field_parser_factory(field_type: str) -> Callable:
    """Manages the parsing of specific fields"""

    mapping = {
        "compound": parse_compound,
        "primitive": parse_primitive,
        "controlledVocabulary": parse_primitive
    }

    return mapping[field_type]


def parse_primitive(field: dict, module):
    """Parses a primtive field by checking the API schema."""

    # Get module properties
    properties = module.schema()["properties"]

    return {
        attr_name: field["value"]
        for attr_name, property in properties.items()
        if property["typeName"] == field["typeName"]
    }


def parse_compound(field: dict, module):
    """Parses a compound field by checking the module schema to a list of JSON objects"""

    # Retrieve the definitions from the module schema
    compound_name, field_mapping = _get_compound_definitions(
        module, field["typeName"])

    return {
        compound_name: [
            {
                field_mapping[type_name]: sub_field["value"]
                for type_name, sub_field in obj.items()
            }
            for obj in field["value"]
        ]
    }


def _get_compound_definitions(module, type_name: str):
    """Retrieves the compound definitions found in the module schema."""

    for obj_name, property in module.schema()["properties"].items():
        compound_name = property["typeName"]

        if compound_name == type_name:

            definition_name = property["items"]["$ref"].split("/")[-1]
            definition = module.schema(
            )["definitions"][definition_name]["properties"]

            # Create a mapping from typeName to actual attribute name
            mapping = {
                field["typeName"]: attr_name
                for attr_name, field in definition.items()
            }

            return obj_name, mapping

    raise NameError(
        f"Field with typeName {type_name} is not defined in module {module.__name__}"
    )



import json
import os
import importlib
import importlib.util
import inspect
import pandas as pd
import sys

from pathlib import Path
from enum import EnumMeta
from tempfile import TemporaryDirectory
from typing import Callable
from datamodel_code_generator import InputFileType, generate
from jinja2 import Template
from importlib import resources as pkg_resources


from easyDataverse.tools import templates as jinja_templates
from easyDataverse.tools.schema import split_metadatablock, generate_JSON_schema, camel_to_snake


def generate_metadatablock_code(path: str, out: str, schema_loc: str) -> None:
    """Generater the code for a metadatablock based on pyDantic"""

    metadatablock_name = pd.read_csv(path, sep="\t", usecols=range(16)).name[0]
    metadatablock_name = metadatablock_name[0].lower(
    ) + metadatablock_name[1::]
    metadatablock = pd.read_csv(path, sep="\t", skiprows=2, usecols=range(16))

    # TODO Fix TSV rather than the columns in the code
    metadatablock.columns = [
        col.strip() for col in metadatablock.columns
    ]

    # Extract all informations
    fields, controlled_vocab = split_metadatablock(metadatablock)
    properties, definitions, required = generate_JSON_schema(
        fields, metadatablock_name, controlled_vocab
    )

    # Generate JSON schema
    filename = f"{metadatablock_name}.json"
    json_schema = {
        "$id": filename,
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": metadatablock_name,
        "type": "object",
        "properties": properties,
        "definitions": definitions,
        "required": required
    }

    schema_path = os.path.join(schema_loc, f"{metadatablock_name}.json")
    with open(schema_path, "w") as f:
        f.write(json.dumps(json_schema, indent=2))

    # Generate the raw code of the metadatablock
    generated_code = generate_code(json_schema)

    # Write code to file to read it afterwards
    module_name = f"{metadatablock_name}.py"
    out_path = os.path.join(out, module_name)

    with open(out_path, "w") as f:
        f.write(generated_code)

    # Convert code to a module to get all sub classes it
    sys.path.append(out)
    code_module = importlib.import_module(metadatablock_name)
    clsmembers = inspect.getmembers(code_module, inspect.isclass)

    for name, obj in clsmembers:
        if name == "DataverseBase":
            continue

        elif obj.__name__.lower() == metadatablock_name.lower():
            continue

        elif isinstance(obj, EnumMeta):
            continue

        generated_code += generate_add_function(obj)

    with open(out_path, "w") as f:
        f.write(clean_code(generated_code))


def generate_code(json_schema: dict) -> str:
    """Generates pyDantic code from a JSON schema."""

    with TemporaryDirectory() as temporary_directory_name:
        temporary_directory = Path(temporary_directory_name)
        output = Path(temporary_directory / 'model.py')
        generate(
            json.dumps(json_schema),
            base_class="easyDataverse.core.DataverseBase",
            input_file_type=InputFileType.JsonSchema,
            input_filename=json_schema["$id"],
            output=output,
            field_include_all_keys=True,
            use_standard_collections=True,
            snake_case_field=True,
            reuse_model=True,
            use_schema_description=True
        )

        return output.read_text()


def clean_code(code: str) -> str:
    """Replaces certain parts of the code, which cannot be enforced via the generator"""

    # Mapping used for replacement
    replace_map = [
        ("...", "default_factory=list"),
        ("type_class", "typeClass"),
        ("type_name", "typeName"),
        ("from core", "from easyDataverse.core")
    ]

    for replacement in replace_map:
        code = code.replace(*replacement)

    return code


def generate_add_function(cls: Callable) -> str:
    """Generates an add function which will be added to the metadatablock's main class."""

    field = camel_to_snake(cls.__name__)
    definitions = generate_definition(cls)
    summary, args_doc = generate_docstring(cls)
    params = generate_arguments(cls)

    template = Template(
        pkg_resources.read_text(jinja_templates, "add_function.jinja2")
    )

    return template.render(
        field=field,
        definitions=definitions,
        summary=summary,
        args_doc=args_doc,
        class_name=cls.__name__,
        params=params
    )


def generate_definition(cls: Callable) -> list[str]:
    """Generates a function signature from a pyDantic class object"""

    # Fetch parameters
    params = cls.__annotations__

    return [
        f"{name}: {data_type},"
        if "Optional" not in data_type else
        f"{name}: {data_type} = None,"
        for name, data_type in params.items()
    ]


def generate_docstring(cls: Callable) -> tuple[str, list[str]]:
    """Generates a Google styled docstring for the add functionality"""

    # Get the properties for all fields
    properties = cls.schema()["properties"]

    # Short generic summary
    summary = f"Function used to add an instance of {cls.__name__} to the metadatablock."

    # Generate individual args lines with descriptions
    args_doc = [
        f"{camel_to_snake(name)} ({field['type']}): {field['description']}"
        if field.get("type") else
        f"{camel_to_snake(name)} (Enum): {field['description']}"
        for name, field in properties.items()
    ]

    # TODO Handle Enums better

    return summary, args_doc


def generate_arguments(cls: Callable) -> str:
    """Generates the arguments that are passed to the class that is about to be instantiated."""

    # Fetch parameters
    params = cls.__signature__.parameters

    return ", ".join([
        f"{camel_to_snake(param)}={camel_to_snake(param)}"
        for param in params.keys()
    ])



import importlib
import glob
import os
import inspect

from pathlib import Path
from jinja2 import Template
from importlib import resources as pkg_resources

from easyDataverse.tools import templates as jinja_templates
from easyDataverse.tools.utils import get_class


def generate_rest_api_code(metadatablock_loc: str, lib_name: str, lib_path: str) -> None:
    """Generates code that will be used to build a REST interface."""

    endpoints, imports = [], []
    # sys.path.append(metadatablock_loc)

    # Initialize Jinja templates
    import_template = Template(
        pkg_resources.read_text(
            jinja_templates, "rest_imports_template.jinja2"
        )
    )
    create_template = Template(
        pkg_resources.read_text(
            jinja_templates, "rest_create_template.jinja2"
        )
    )

    for path in glob.glob(os.path.join(metadatablock_loc, "*.py")):

        # Get module name
        module_name = Path(path).stem

        if module_name == "__init__":
            continue

        # Get the class name
        class_name, _ = get_class(module_name, path)

        endpoints.append(create_template.render(
            module_name=module_name, class_name=class_name
        ))

        # Add infos to imports
        imports.append({
            "module": module_name, "class": class_name
        })

    # Finally, write REST API to file
    out_path = os.path.join(lib_path, f"{lib_name.lower()}_server.py")
    with open(out_path, "w") as f:
        f.write(import_template.render(imports=imports, lib_name="pyDaRUS"))
        f.write("\n\n".join(endpoints))


def get_class_name(module_name: str, module_path: str) -> str:
    """Retrieves the correct class name of a module"""

    # Import the module
    module = importlib.import_module(module_name, module_path)

    for class_name, _ in inspect.getmembers(module, inspect.isclass):
        if class_name.lower() == module_name.lower():
            return class_name

    raise ModuleNotFoundError(
        f"Couldnt find corresponding class {module_name} in {module_path}."
    )



import argparse
import os
import glob
import yaml
import json
import logging
import coloredlogs


from jinja2 import Template
from typing import Callable
from importlib import resources as pkg_resources

from easyDataverse.tools import templates as jinja_templates
from easyDataverse.tools.clsmod import generate_metadatablock_code
from easyDataverse.tools.restmod import generate_rest_api_code
from easyDataverse.tools.filehandling import create_init_line_metadatblock, create_init_line_library, generate_template


class MyDumper(yaml.Dumper):
    """Used to fix the bad indentation from pyaml"""

    def increase_indent(self, flow=False, indentless=False):
        return super(MyDumper, self).increase_indent(flow, False)


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--path", required=True,
                        type=str, help="Path to the TSV files describing metadatablocks.")
    parser.add_argument("-o", "--out", required=True,
                        type=str, help="Path to which the API will be written.")
    parser.add_argument("-n", "--name", required=True,
                        type=str, help="Name of the generated API.")
    parser.add_argument("-l", "--logger", required=False, default="INFO",
                        type=str, help="Name of the generated API.")
    args = vars(parser.parse_args())

    # Logger
    global logger
    logger = logging.getLogger(__name__)
    coloredlogs.install(level=args["logger"])

    # Parse args to global variables
    path = os.path.abspath(os.path.join(args["path"]))
    out = os.path.abspath(os.path.join(args["out"]))
    name = args["name"]

    # Process args
    project_path = os.path.join(out, name)
    lib_path = os.path.join(out, name, name)
    metadatablock_loc = os.path.join(out, name, name, "metadatablocks")
    schema_loc = os.path.join(out, name, name, "metadatablocks", "schemas")
    template_loc = os.path.join(out, name, name, "templates")

    # Create dir structure
    os.makedirs(schema_loc, exist_ok=True)
    os.makedirs(os.path.join(template_loc, "yaml"), exist_ok=True)
    os.makedirs(os.path.join(template_loc, "json"), exist_ok=True)

    # Create metadatablock code
    metadatablocks(path, metadatablock_loc, schema_loc, lib_path, name)

    # Create setup file
    setup(name, project_path)

    # Create templates
    templates(metadatablock_loc, template_loc)

    # Generate REST API
    generate_rest_api_code(metadatablock_loc, name, lib_path)

    logger.info(
        f"Created REST-API for {name}"
    )

    # Generate Dockerfile
    docker_template = Template(
        pkg_resources.read_text(
            jinja_templates, "dockerfile.jinja2"
        )
    )

    docker_out = os.path.join(project_path, "Dockerfile")
    with open(docker_out, "w") as f:
        f.write(
            docker_template.render(lib_name=name)
        )

    logger.info(
        f"Created library {name} in {os.path.abspath(project_path)}"
    )


def metadatablocks(path: str, metadatablock_loc: str, schema_loc: str, lib_loc: str, lib_name: str) -> None:
    """Generates the metadatablock relevant files for the API."""

    # Generate code for the metadatablocks
    for block_path in glob.glob(os.path.join(path, "*.tsv")):
        generate_metadatablock_code(block_path, metadatablock_loc, schema_loc)
        logger.info(f"Generated metadatablock code for {block_path}")

    # Get the correspding module names form the files
    module_search = os.path.join(metadatablock_loc, "*.py")

    # Write __init__ files
    write_imports(
        module_search,
        lib_loc,
        "library_init.jinja2",
        create_init_line_library,
        lib_name
    )

    write_imports(
        module_search,
        metadatablock_loc,
        "metadatablock_init.jinja2",
        create_init_line_metadatblock,
        lib_name
    )


def write_imports(
    module_search: str,
    path: str,
    template_path: str,
    fun: Callable,
    lib_name: str
):
    """Extract module names, creates imports via a function and writes them to a path"""

    imports = [
        fun(module, lib_name)
        for module in glob.glob(module_search)
        if "__init__" not in module
    ]

    with open(os.path.join(path, "__init__.py"), "w") as f:
        template = Template(
            pkg_resources.read_text(jinja_templates, template_path)
        )

        f.write(template.render(imports=imports, lib_name=lib_name))


def setup(name: str, project_path: str) -> None:
    """Generates the relevant setup file to install the API."""

    requirements = [
        "easyDataverse",
        "fastapi",
        "uvicorn",
        "pydantic",
        "jinja2",
        "pyDataverse",
        "pandas",
        "pyaml"
    ]

    # Initialize Jinja template
    template = Template(
        pkg_resources.read_text(jinja_templates, "setup.jinja2")
    )

    # Write to the API directory
    setup_path = os.path.join(project_path, "setup.py")
    with open(setup_path, "w") as f:
        f.write(template.render(name=name, requirements=requirements))

    # Write requirements.txt
    requirements_path = os.path.join(project_path, "requirements.txt")
    with open(requirements_path, "w") as f:
        for req in requirements:
            f.write(req + "\n")


def templates(metadatablock_loc: str, template_loc: str) -> None:
    """Generates templates which can then be used for mapping from file formats to Dataverse."""

    # Get the correspding module names form the files
    module_search = os.path.join(metadatablock_loc, "*.py")

    for module in glob.glob(module_search):

        if "__init__" in module:
            continue

        # Generate empty data model
        block_name, data_model = generate_template(module)

        # Write template to YAML
        with open(os.path.join(template_loc, "yaml", f"{block_name}.yaml"), "w") as f:
            yaml.dump(
                data_model, f, sort_keys=False, default_flow_style=False, Dumper=MyDumper
            )

        # Write template to JSON
        with open(os.path.join(template_loc, "json", f"{block_name}.json"), "w") as f:
            json.dump(data_model, f, indent=2, sort_keys=False)

        logger.info(
            f"Created template for {block_name}."
        )


if __name__ == "__main__":
    main()



import re
import logging
import pandas as pd

# Typing that is used to infer proper types
type_mapping = {
    "text": "string",
    "url": "string",
    "float": "number",  # TODO Float not parsed as float rather as number
    "integer": "integer",
    "int": "integer",
    "textbox": "string",
    "date": "string",
    "email": "string"
}


def camel_to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


def spaced_to_snake(name) -> str:

    # Clean the title
    name = re.sub(r"-|\?",  '', name)
    name = name.replace(r'/', ' ')

    return "_".join([
        word.lower()
        for word in name.replace("-", " ").split(" ")
    ])


def split_metadatablock(metadatablock: pd.DataFrame):
    """Splits a Metadatablock to the fields and controlled vocabularies if given."""

    split_index: int = metadatablock.index[
        metadatablock["#datasetField"] == "#controlledVocabulary"
    ].tolist()[0]

    # Get fields and vocabulary
    fields = metadatablock.iloc[0:split_index]
    controlled_vocab = metadatablock.iloc[split_index::]

    # Set top row as columns for controlled vocab
    controlled_vocab.columns = controlled_vocab.iloc[0]
    controlled_vocab = controlled_vocab[1::]

    return fields, controlled_vocab


def generate_JSON_schema(fields: pd.DataFrame, metadatablock_name: str, controlled_vocab: pd.DataFrame):
    """Sorts Fields by compounds or single fields"""

    # Initialize compound and single field dictionaries
    properties = {}
    definitions = {}
    required = []

    # Get all parent fields
    parents = set(fields.parent)

    for parent in parents:

        if str(parent) == "nan":

            # Process any primitive field to JSON
            primitives = fields[(fields["parent"].isna())
                                & (fields["fieldType"] != "none")]
            property = fetch_fields(
                primitives.to_dict(orient="records"),
                controlled_vocab
            )

        else:
            # Process data to a compound JSON
            property, compound_name, definition = construct_compound(
                parent, fields, controlled_vocab
            )

            # Add to Schema
            required.append(compound_name)
            definitions.update(definition)

        properties.update(property)

    # Add property to identify class as a metadatablock
    properties.update({
        "_metadatablock_name": {
            "default": metadatablock_name,
            "type": "string"
        }
    })

    return properties, definitions, required


def construct_compound(
    parent: str,
    fields: pd.DataFrame,
    controlled_vocab: pd.DataFrame
):
    """Returns the JSON schema properties as well as definitions"""

    # Get Compound meta information
    compound_row = fields[fields.name == parent]
    compound_dict = compound_row.to_dict(orient="records")[0]
    compound_name = spaced_to_snake(compound_dict["title"])

    # Fetch all fields corresponding to the compound
    fields = fields[fields.parent == parent].to_dict(orient="records")

    # Get Compound schema for property
    object_props = {
        compound_name: {
            "typeName": compound_dict['name'],
            "type": "array",
            "typeClass": "compound",
            "multiple": bool(compound_dict["allowmultiples"]),
            "description": compound_dict["description"],
            "items": {
                "$ref": f"#/definitions/{compound_name}"
            }
        }
    }

    fields = fetch_fields(fields, controlled_vocab)

    return object_props, compound_name, {
        compound_name: {
            "properties": fields
        }
    }


def fetch_fields(fields: list[dict], controlled_vocab: pd.DataFrame):
    """Retrieves and parses fields given in a metadatablock"""

    # Initialize field dictionary
    field_defs = {}

    for field in fields:
        field_defs.update(
            parse_field(field, controlled_vocab)
        )

    return field_defs


def parse_field(field: dict, controlled_vocab: pd.DataFrame):
    """Parses a field to the appropriate JSON schema and retrieves controlled_vocabs"""

    # Set defaults
    field_name = spaced_to_snake(field["title"])
    type_class = "primitive"
    allowed_vals = None

    if field["allowmultiples"]:
        data_type = "array"
    else:
        data_type = type_mapping[field["fieldType"]]

    if field["allowControlledVocabulary"]:
        type_class = "controlledVocabulary"
        allowed_vals = get_controlled_vocabs(field["name"], controlled_vocab)

    field_def = {
        "typeName": field["name"],
        "typeClass": type_class,
        "multiple": field["allowmultiples"],
        "type": data_type,
        "description": field["description"]
    }

    if allowed_vals:
        field_def["enum"] = allowed_vals

    if field_name == "subject":
        # TODO find elegant way to handle special fields
        field_def = {
            "typeName": field["name"],
            "typeClass": type_class,
            "multiple": field["allowmultiples"],
            "type": "array",
            "items": {
                "type": "string",
                "enum": allowed_vals
            },
            "description": field["description"]
        }

    return {
        field_name: field_def
    }


def get_controlled_vocabs(name: str, controlled_vocab: pd.DataFrame):
    """Retrieves the possible values for a controlled vocabulary"""

    cv_fields = controlled_vocab[controlled_vocab["DatasetField"] == name]
    return cv_fields["Value"].tolist()



import os
import shutil

from typing import List, Optional

from pyDataverse.api import NativeApi
from pyDataverse.models import Dataset, Datafile
from easyDataverse.core.exceptions import MissingURLException, MissingCredentialsException


def upload_to_dataverse(
    json_data: str,
    dataverse_name: str,
    filenames: List[str] = None,
    p_id: str = None,
    DATAVERSE_URL: Optional[str] = None,
    API_TOKEN: Optional[str] = None
) -> str:
    """Uploads a given Dataset to the dataverse installation found in the environment variables.

    Args:
        json_data (str): JSON representation of the Dataverse dataset.
        dataverse_name (str): Name of the Dataverse where the data will be uploaded to.
        filenames (List[str], optional): List of files that should be uploaded. Can also include durectory names. Defaults to None.

    Raises:
        MissingURLException: URL to the dataverse installation is missing. Please include in your environment variables.
        MissingCredentialsException: API-Token to the dataverse installation is missing. Please include in your environment variables

    Returns:
        str: The resulting DOI of the dataset, if successful.
    """

    # Get environment variables
    if DATAVERSE_URL is None:
        try:
            DATAVERSE_URL = os.environ["DATAVERSE_URL"]

        except KeyError:
            raise MissingURLException

    if API_TOKEN is None:
        try:
            API_TOKEN = os.environ["DATAVERSE_API_TOKEN"]
        except KeyError:
            raise MissingCredentialsException

    # Initialize pyDataverse API and Dataset
    api = NativeApi(DATAVERSE_URL, API_TOKEN)
    ds = Dataset()
    ds.from_json(json_data)

    # Finally, validate the JSON
    if ds.validate_json():

        if p_id:
            # Update dataset if pid given
            response = api.create_dataset(
                dataverse_name, json_data, p_id
            )
        else:
            # Create new if no pid given
            response = api.create_dataset(
                dataverse_name, json_data
            )

        if response.json()["status"] != "OK":
            raise Exception(response.json()["message"])

        # Get response data
        ds_pid = response.json()["data"]["persistentId"]

        if filenames is not None:
            # Upload files if given
            for filename in filenames:
                __uploadFile(
                    filename=filename,
                    ds_pid=ds_pid,
                    api=api
                )

        return ds_pid

    else:
        raise Exception("Could not upload")


def __uploadFile(filename: str, ds_pid: str, api: NativeApi) -> None:
    """Uploads any file to a dataverse dataset.
    Args:
        filename (String): Path to the file
        ds_pid (String): Dataset permanent ID to upload.
        api (API): API object which is used to upload the file
    """

    # Check if its a dir
    if os.path.isdir(filename):
        shutil.make_archive("contents", 'zip', filename)
        filename = "contents.zip"

    df = Datafile()
    df.set({"pid": ds_pid, "filename": filename})
    api.upload_datafile(ds_pid, filename, df.json())



import json
import yaml
import xmltodict

from pydantic import BaseModel


class DataverseBase(BaseModel):

    class Config:
        validate_all = True
        validate_assignment = True
        use_enum_values = True

    @classmethod
    def from_json_string(cls, json_string: str):
        """Initializes an object from a JSON file"""

        return cls.parse_obj(
            json.loads(json_string)
        )

    @classmethod
    def from_json_file(cls, file_path: str):
        """Initializes an object from a JSON file"""

        with open(file_path, 'r') as f:
            return cls.parse_obj(
                json.load(f)
            )

    @classmethod
    def from_yaml_string(cls, yaml_string: str):
        """Initializes an object from a YAML string"""

        return cls.parse_obj(
            yaml.safe_load(yaml_string)
        )

    @classmethod
    def from_yaml_file(cls, file_path: str):
        """Initializes an object from a YAML string"""

        with open(file_path, 'r') as f:
            return cls.parse_obj(
                yaml.safe_load(f)
            )

    def json(self, indent: int = 2, **kwargs) -> str:
        """Returns a JSON representation of the dataverse object."""

        # Read the JSON to filter empty compounds
        json_obj = json.loads(super().json(
            exclude_none=True, indent=indent, **kwargs))

        return json.dumps({
            key: value
            for key, value in json_obj.items()
            if value != []
        }, indent=indent)

    def dict(self, **dictkwargs) -> dict:
        """Returns a dictionary representation of the dataverse object."""

        # Get the dictionary function from pyDantic
        fields = super().dict(**dictkwargs)

        return {
            key: value
            for key, value in fields.items()
            if value != []
        }

    def xml(self, **dictkwargs) -> str:
        """Returns an XML representation of the dataverse object."""

        # Turn all fields to camel case
        fields = self._keys_to_camel(
            {self.__class__.__name__: self.dict(**dictkwargs)}
        )

        return xmltodict.unparse(
            fields,
            pretty=True, indent="    "
        )

    def _keys_to_camel(self, dictionary: dict):
        nu_dict = {}
        for key in dictionary.keys():
            if isinstance(dictionary[key], dict):
                nu_dict[
                    self._snake_to_camel(key)
                ] = self._keys_to_camel(dictionary[key])
            else:
                nu_dict[
                    self._snake_to_camel(key)
                ] = dictionary[key]
        return nu_dict

    @staticmethod
    def _snake_to_camel(word: str) -> str:
        return ''.join(x.capitalize() or '_' for x in word.split('_'))

    def dataverse_dict(self) -> dict:
        """Converts a metadatablock object model to the appropriate dataverse JSON format"""

        # Get properties and init json_obj
        schema_props = self.schema()["properties"]
        json_obj = {}

        for attr, value in self.__dict__.items():

            if value == []:
                # Guard clause to catch empty compounds
                continue

            # Fetch dataverse infos
            properties = schema_props[attr]

            # Process compounds
            if properties["typeClass"] == "compound":
                value = [
                    field.dataverse_dict()
                    for field in value
                ]

            # Assign everything
            if value:
                json_obj.update({
                    properties["typeName"]: {
                        "multiple": properties["multiple"],
                        "typeClass": properties["typeClass"],
                        "typeName": properties["typeName"],
                        "value": value if isinstance(value, list) else str(value)
                    }
                })

        if hasattr(self, "_metadatablock_name"):
            return {
                getattr(self, "_metadatablock_name"): {
                    "fields": list(json_obj.values())
                }
            }
        else:
            return json_obj

    def to_dataverse_json(self, indent: int = 2) -> str:
        """Returns a JSON formatted representation of the dataverse object."""
        return json.dumps(self.dataverse_dict(), indent=indent)



import setuptools
from setuptools import setup

setup(
    name='easyDataverse',
    version='0.2.7',
    author='Jan Range',
    author_email='jan.range@simtech.uni-stuttgart.de',
    license='MIT License',
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': ['api-generator=easyDataverse.api_generator:main']
    },
    install_requires=[
        'pydantic',
        'jinja2',
        'pydataverse',
        'pandas',
        'datamodel_code_generator',
        'pyaml',
        'coloredlogs',
        'xmltodict'
    ]
)



from typing import Any
from operator import itemgetter
from lxml import objectify, etree

from sdRDM.tools.utils import snake_to_camel


class XMLWriter:

    @classmethod
    def to_xml(
        cls,
        root_name: str,
        values: dict,
        props: dict,
        to_string: bool = True
    ) -> objectify.ObjectifiedElement:
        """Writes the object data model to XML"""

        # Initialize attrib/element mapping
        attributes = {}
        elements = []

        # Collect attributes and elements
        for name, value in values.items():
            if name.startswith("__"):
                continue

            # Get xml class and data type
            xml, data_type = cls._get_xml_specs(props[name])

            # Treat all those cases
            if xml == "attribute":
                attributes.update({name: str(value)})

            elif xml == "element":
                elements.append(cls._make_element(
                    name, value, data_type
                ))

        # Construct resulting element
        root = objectify.Element(
            root_name,
            **attributes
        )

        root.extend(elements)

        # Some cleanups
        objectify.deannotate(root)
        etree.cleanup_namespaces(root)

        if to_string:
            xml_string = etree.tostring(
                root,
                pretty_print=True,
                xml_declaration=True
            )

            return xml_string.decode()

        return root

    @staticmethod
    def _get_xml_specs(properties: dict) -> tuple[str, str]:
        """Extracts the xml classification and type"""

        def infer_property(key: str):
            mapping = {
                "xml": "element",  # Make it an element if not otherwise stated
                "type": "object"
            }

            try:
                return properties[key]
            except KeyError:
                return mapping[key]

        return (
            infer_property("xml"),
            infer_property("type")
        )

    @staticmethod
    def _make_element(name: str, value: Any, data_type: str):
        """Creates elements based on their type and recursively
           generated elements from other classes"""

        name = snake_to_camel(name)

        if data_type not in ["object", "array"]:
            # Primitive types
            elem = etree.Element(name)
            elem.text = str(value)
            return elem

        # Process nested types
        elem = objectify.Element(name)

        if data_type == "object":
            elem.extend([value.to_xml(to_string=False)])
        elif data_type == "array":
            elem.extend([entry.to_xml(to_string=False) for entry in value])
        else:
            raise TypeError(f"Data type of {data_type} is unknown.")

        return elem



import os
import xmltodict

from typing import Dict, Any, List, Optional
from pydantic import BaseModel, validate_arguments
from json import dumps

from easyDataverse.core.uploader import upload_to_dataverse
from easyDataverse.core.dataverseBase import DataverseBase
from easyDataverse.core.downloader import download_from_dataverse
from easyDataverse.core.exceptions import DatasetAlreadyExistsException


class Dataset(BaseModel):

    class Config:
        extra = "allow"

    metadatablocks: Dict[str, Any] = dict()
    p_id: Optional[str] = None

    # ! Adders

    def add_metadatablock(
        self,
        metadatablock: DataverseBase
    ) -> None:
        """Adds a metadatablock object to the dataset if it is of 'DataverseBase' type and has a metadatablock name"""

        # Check if the metadatablock is of 'DataverseBase' type
        if issubclass(metadatablock.__class__, DataverseBase) is False:
            raise TypeError(
                f"Expected class of type 'DataverseBase', got '{metadatablock.__class__.__name__}'"
            )

        if hasattr(metadatablock, '_metadatablock_name') is False:
            raise TypeError(
                f"The provided class {metadatablock.__class__.__name__} has no metadatablock name and is thus not compatible with this function."
            )

        # Add the metadatablock to the dataset as a dict
        block_name = getattr(metadatablock, "_metadatablock_name")
        self.metadatablocks.update(
            {block_name: metadatablock}
        )

        # ... and to the __dict__
        setattr(self, block_name, metadatablock)

    # ! Exporters

    def xml(self) -> str:
        """Returns an XML representation of the dataverse object."""

        # Turn all keys to be camelcase
        fields = self._keys_to_camel(
            {"dataset_version": self.dict()}
        )

        return xmltodict.unparse(
            fields,
            pretty=True, indent="    "
        )

    def dataverse_dict(self) -> dict:
        """Returns a dictionary representation of the dataverse dataset."""

        # Convert all blocks to the appropriate format
        blocks = {}
        for block in self.metadatablocks.values():
            blocks.update(block.dataverse_dict())

        return {
            "datasetVersion": {
                "metadataBlocks": blocks
            }
        }

    def dataverse_json(self, indent: int = 2) -> str:
        """Returns a JSON representation of the dataverse dataset."""

        return dumps(self.dataverse_dict(), indent=indent)

    # ! Dataverse interfaces

    def upload(self, dataverse_name: str, filenames: List[str] = None, base_url: Optional[str] = None, api_token: Optional[str] = None) -> str:
        """Uploads a given dataset to a Dataverse installation specified in the environment variable.

        Args:
            dataverse_name (str): Name of the target dataverse.
            filenames (List[str], optional): File or directory names which will be uploaded. Defaults to None.

        """

        self.p_id = upload_to_dataverse(
            json_data=self.dataverse_json(),
            dataverse_name=dataverse_name,
            filenames=filenames,
            p_id=self.p_id,
            DATAVERSE_URL=base_url,
            API_TOKEN=api_token
        )

        return self.p_id

    # ! Initializers

    @classmethod
    @validate_arguments
    def from_dataverse_doi(cls, doi: str):
        """Initializes a Dataset from a given Dataverse dataset.

        Args:
            doi (str): DOI of the dataverse Dataset.
            lib_name (str): Name of the library to fetch the given metadatablocks.

        Returns:
            Dataset: Resulting dataset that has been downloaded from Dataverse.
        """

        lib_name = os.environ["EASYDATAVERSE_LIB_NAME"]

        return download_from_dataverse(cls(), doi, lib_name)

    # ! Utilities

    @staticmethod
    def _snake_to_camel(word: str) -> str:
        return ''.join(x.capitalize() or '_' for x in word.split('_'))

    def _keys_to_camel(self, dictionary: dict):
        nu_dict = {}
        for key in dictionary.keys():
            if isinstance(dictionary[key], dict):
                nu_dict[
                    self._snake_to_camel(key)
                ] = self._keys_to_camel(dictionary[key])
            else:
                nu_dict[
                    self._snake_to_camel(key)
                ] = dictionary[key]
        return nu_dict



class ControlledVocabularyError(Exception):
    """Raises an error if a value provided by a controlled vocabulary doesnt match"""
    pass


class MissingCredentialsException(Exception):
    """Throws an exception if credentials are not provided in the environment variables"""

    def __init__(self) -> None:
        self.message = (
            "Please specify your API Token in your environment variables with the key 'DATAVERSE_API_TOKEN'"
        )

        super().__init__(self.message)


class MissingURLException(Exception):
    """Throws an exception if no dataverse URL has been provided in the environment variables"""

    def __init__(self) -> None:
        self.message = (
            "Please specify a Dataverse instance URL in your environment variables with the key 'DATAVERSE_URL'"
        )

        super().__init__(self.message)


class DatasetAlreadyExistsException(Exception):
    """Throws an exception if the dataset has already been uploaded --> Existing p_id in Dataset"""

    def __init__(self, p_id) -> None:
        self.message = (
            f"The dataset has already been uploaded to your Dataverse, since the Dataset object has a given 'p_id' of {p_id}. \
            This error might be the cause of double execution of the Dataset's upload method. However, if you want to update your \
            dataset, use the Dataset's update method."
        )

        super().__init__(self.message)


class ValidationError(Exception):
    """Throws an exception if validation failed and prints the error message"""
    pass



import inspect
import importlib
import importlib.util
import sys

from pathlib import Path


def create_init_line_library(module_path: str, lib_name: str) -> str:
    """Creates an __init__ entry of the library dir for the corresponding metadatablocks file"""

    # Get the module name from file
    module_name = Path(module_path).stem

    # Import package to inspect it
    module = importlib.import_module(module_name)

    # Get the class that will be part of the init
    for class_name, _ in inspect.getmembers(module, inspect.isclass):

        if class_name.lower() == module_name.lower():
            return (
                f"from {lib_name}.metadatablocks import {class_name}"
            )

    raise ModuleNotFoundError(
        f"Could not find module '{module_name}' in generated modules."
    )


def create_init_line_metadatblock(module_path: str, lib_name: str) -> str:
    """Creates an __init__ entry of the metadatablocks dir for the corresponding metadatablocks file"""

    # Get the module name from file
    module_name = Path(module_path).stem

    # Import package to inspect it
    module = importlib.import_module(module_name, module_path)

    # Get the class that will be part of the init
    for class_name, _ in inspect.getmembers(module, inspect.isclass):

        if class_name.lower() == module_name.lower():
            return (
                f"from {lib_name}.metadatablocks.{module_name} import {class_name}"
            )

    raise ModuleNotFoundError(
        f"Could not find module '{module_name}' in generated modules."
    )


def generate_template(module_path: str) -> tuple[str, dict]:
    """Generates templates which can be used to map from arbitrary file formats to Dataverse"""

    # Get the module name from file
    module_name = Path(module_path).stem

    # Import package to inspect it
    module = importlib.import_module(module_name, module_path)

    # Get the class that will be part of the init
    for class_name, class_obj in inspect.getmembers(module, inspect.isclass):
        if class_name.lower() == module_name.lower():
            return class_name, {
                class_name: class_to_empty_dict(class_obj)
            }

    raise ModuleNotFoundError(
        f"Could not find module '{module_name}' in generated modules."
    )


def class_to_empty_dict(cls) -> dict:
    """Generates a dictionary with 'None' fields which can be used for mapping."""

    data_model, array_fields, single_fields = {}, {}, {}
    for field_name, field in cls.__fields__.items():
        try:
            # Parse compound fields
            subclass_properties = field.type_.schema()["properties"]
            array_fields.update(
                {field_name: [{
                    attribute: None
                    for attribute in subclass_properties.keys()
                }]
                }
            )
        except AttributeError:
            # Parse primitive fields
            single_fields.update(
                {field_name: None}
            )

    # Merge both to have single fields on top and array fields bottom
    data_model.update(single_fields)
    data_model.update(array_fields)

    return data_model
